==> excc-0.1.2.tar.gz: excc-0.1.2/exception.h <==
/* exception.h --- header implementing lightweight, but featured
                   exception handling.

Time-stamp: "2004-01-30 23:48:12 admp"

Copyright (C) 2003 admp <admp@licejus.lt>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA. */

/* Summary.

To understand all this, you should be experienced in C, and also
know the setjmp interface.  (Unixers do.)

Implemented:
  1. TRY --- a catcher block.
  2. EXCEPT, EXPECT, CATCH --- a handler block.
  3. RAISE --- raises an exception.
  4. ON --- a handler.

IMPORTANT:
  TRY without EXCEPT does not work properly.  This is impossible
  to fix, at least I don't see the way.  Anyway, if you don't
  want to break the whole system, I see no other need for
  such a construct.

Syntax:
  1. TRY
  The syntax is:
    try { statements; } except { handler blocks; }

  2. EXCEPT, EXPECT, CATCH
  The syntax is:
    except { handler blocks; }

  3. RAISE
  The syntax is: 
    raise (params)
  PARAMS depend on your local implementation.  By default, it is a
  number (int).
  
  4. ON
  The syntax is:
    on (params) { statements; }
  PARAMS are the same as for RAISE.

Supports:
  1. Nested catcher blocks.
  2. Raising an exception outside catcher block (this results in an 
     unhandled exception.)
  3. Raising an exception in function called from catcher block.
  4. Different types for exception structures.
  5. Reporting file and line where exception was raised/handled. */

/* Search for TODO and FIXME tags if you want to enchance this.
   (This would be nice.) */

/* TODO: since customized exceptions usually allocate memory
   dynamically, it would be nice to have support for freeing
   function, which would be called after it's not needed
   anymore: __EXC_FREE.  I. e. after exception is handled. */

#if defined __EXCEPTION_H__
  /* TODO: allow compile as whole-in-one, header OR source file.
     This is the most wanted feature - currently this header
     can be used in single file applications [wrong]. */
#  error Header file "exception.h" used twice.  This is the Wrong Thing.
#endif

#define __EXCEPTION_H__

#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* The default prefix is `__exc' for functions and `__EXC' for
   macros.  Underscores mean you shouldn't touch 'em! */



/* C syntax hacks.

   Oh, god, I felt like an inventor after writing these.
   Clearly, code

     for (start (), J = 1; J; end (), J = 0)
       code ();

   Does this:
   1) Executes START
   2) Executes CODE
   3) Executes END
   4) ...And terminates.

   It also works if nested (think why yourself.)
*/

/* Execute START, then block after the macro, and finally END. */

#define __EXC_BLOCK(start, end)              \
  for (start, __exc_block_pass = 1;          \
       __exc_block_pass;                     \
       end, __exc_block_pass = 0)

/* Likewise, but START block is empty. */

#define __EXC_END(end)                       \
  for (__exc_block_pass = 1;                 \
       __exc_block_pass;                     \
       end, __exc_block_pass = 0)



/* For function name.  GCC includes things which expand to
   the name of current function's name.  */

#if (!defined (__GNUC__) || __GNUC__ < 2 || \
     __GNUC_MINOR__ < (defined (__cplusplus) ? 6 : 4))
   /* Otherwise stick to unknown. */
#  define __EXC_FUNCTION               (char *) 0
#else
#  define __EXC_FUNCTION	       __PRETTY_FUNCTION__
#endif



/* The exception value. */

/* You'll want to make local changes to these.  For example, to use
   your own exception structure. */

/* Exception is by default an int.  Anyway, it can be anything from
   string to some structure.

   Whatever the implementation you choose, type name should be
   defined as __EXC_TYPE.  The RAISE (and ON) macro accepts as
   many arguments, s it is given, so your function may use all
   the power of argument passing.  Define your function's name
   as __EXC_MAKE. Exceptions are compared in ON macro.  You
   should define comparing function as __EXC_EQ.

   For example, if you'd like to use strings in place of numbers,
   use this snippet:

   1) #define __EXC_TYPE         char *
   2) #define __EXC_EQ(s1, s2)   (strcasecmp (s1, s2) == 0)
   3) #define __EXC_PRINT(e, stream) \
	         fprintf (stream, "%s", e)

*/

#ifndef __EXC_TYPE
#  define __EXC_TYPE               int

/* Include the default __EXC_PRINT. */
#  define __EXC_TYPE_DEFAULT
#endif

#ifndef __EXC_MAKE
#  define __EXC_MAKE(code...)      code
#endif

#ifndef __EXC_EQ
#  define __EXC_EQ(c1, c2)         ((c1) == (c2))
#endif


/* Optional exception printer.  This is used for debugging purposes
   only.  Define yourself's one as __EXC_PRINT. Arguments are
   exception of type __EXC_TYPE and stream to print to. */

#if !defined (__EXC_PRINT) && defined (__EXC_TYPE_DEFAULT)
#  define __EXC_PRINT(e, stream)             \
     fprintf (stream, "%d", e)
#endif



/* All variables are declared volatile to force non-optimization. */

/* This counter is used by __EXC_BLOCK.  It works well even if nested. */
static volatile int __exc_block_pass;

/* Flag to be set by ON? */
static volatile int __exc_handled;

/* For indexing every call to TRY. */
static volatile unsigned __exc_tries;

/* These identify the raised exception.  File, function, line and
   the exception itself. */
static char *__exc_file;
static char *__exc_function;
static unsigned __exc_line;
static __EXC_TYPE __exc_code;

/* Stack is actually a linked list of catcher cells. */
struct __exc_stack
{
  unsigned num;
  jmp_buf j;
  struct __exc_stack *prev;
};

/* This is the global stack of catchers. */
static struct __exc_stack *__exc_global;



/* Debugging of exceptions.  Nothing interesting for you.  (Just for me.) 
   Anyway, it generates many (really) messages telling what is going
   to happen.  I order to work with it successfully, you should define
   __EXC_PRINT (see above.)
*/
#ifdef __EXC_DEBUG
#  include <stdarg.h>
#  ifndef __EXC_STREAM
/* I often redirect debugging information to a file. */
#    define __EXC_STREAM                     stdout
#  endif

/* Prints error message. */
static void
__exc_debug (char *fmt, ...)
{
  va_list ap;

  fprintf (__EXC_STREAM, "__EXC: ");
  va_start (ap, fmt);
  vfprintf (__EXC_STREAM, fmt, ap);
  va_end (ap);
  fprintf (__EXC_STREAM, "\n");
}

/* For printing __exc_global. */

static void
__exc_print_global ()
{
  struct __exc_stack *level = __exc_global;
  unsigned items;

  if (level == NULL)
    {
      fprintf (__EXC_STREAM, "Stack empty\n");
      return;
    }

  fprintf (__EXC_STREAM, "Current stack (from bottom to top):\n");
  for (items = 0; level; level = level->prev)
    {
      fprintf (__EXC_STREAM, "%c ", items == 0 ? '[' : ' ');

      /* FIXME: does printing jmp_buf literally (as we now address and
	 its size) good?  No, unless human understands it.  One
	 can't unless he chooses one popular C library and study the
	 implementation carefully;) */
      fprintf (__EXC_STREAM, "%u", level->num);
      /* TODO: newline every four or so items.  Use tab stops. */
      fprintf (__EXC_STREAM, " %c\n", level->prev ? ' ' : ']');  
      items++;
    }

  fprintf (__EXC_STREAM, "Totally %u items.\n", items);
}

#else
#  define __exc_debug(args...)
#  define __exc_print_global()
#endif



/* Prints information about exception.  Called in debug mode, or when
   no handler is found. */

void
__exc_print (FILE *stream, char *file, char *function, unsigned line,
	     __EXC_TYPE code)
{
  fprintf (stream, "Exception in file \"%s\", at line %u",
	   file, line);
  if (function)
    {
      fprintf (stream, ", in function \"%s\"", function);
    }
  fprintf (stream, ".");

#ifdef __EXC_PRINT
  fprintf (stream, " Exception: ");
  __EXC_PRINT (code, stream);
#endif
  fprintf (stream, "\n");
}



/* Pop exception from stack, putting into J (if nonzero).  If stack is
   empty, print error message and exit.  Used in EXCEPT. */

static void
__exc_pop (jmp_buf *j)
{
  register struct __exc_stack *stored = __exc_global;

  __exc_debug ("POP () to %p", j);

  if (stored == NULL)
    {
      __exc_debug ("Unhandled exception.");

      fprintf (stderr, "Unhandled exception:\n");
      __exc_print (stderr, __exc_file, __exc_function,
		   __exc_line, __exc_code);

      exit (3);
    }

  __exc_global = stored->prev;

  if (j)
    {
      /* This assumes that JMP_BUF is a structure etc. and can be
	 copied rawely.  This is true in GLIBC, as I know. */
      memcpy (j, &stored->j, sizeof (jmp_buf));
    }

  __exc_debug ("Popped");
  __exc_print_global ();

  /* While with MALLOC, free.  When using obstacks it is better not to
     free and hold up. */
  free (stored);
}



/* Push J onto the stack, with RETURNED as value from SETJMP.  Return
   nonzero, if RETURNED is 0.  If RETURNED is nonzero, returns 0.
   Used in TRY. */

static int
__exc_push (jmp_buf *j, int returned)
{
  struct __exc_stack *new;

  __exc_debug ("PUSH (), %p, %d", j, returned);

  /* SETJMP returns 0 first time, nonzero from __EXC_RAISE.
     Returning false-like value here (0) will enter the
     else branch (that is, EXCEPT.) */
  if (returned != 0)
    {
      __exc_debug ("Returning from RAISE");
      return 0;
    }

  /* Since this didn't come from RAISE, fine to increase counter. */
  ++__exc_tries;
  __exc_debug ("This is PUSH () number %u", __exc_tries);

  /* Using memcpy here is the best alternative. */
  new = malloc (sizeof (struct __exc_stack));
  memcpy (&new->j, j, sizeof (jmp_buf));
  new->num = __exc_tries;
  new->prev = __exc_global;
  __exc_global = new;

  __exc_print_global ();

  return 1;
}



/* Raise an exception in FILE at LINE, with code CODE.  Used in RAISE. */

static void
__exc_raise (char *file, char *function, unsigned line, __EXC_TYPE code)
{
  jmp_buf j;

  __exc_debug ("RAISE ()");
#if defined __EXC_DEBUG
  __exc_print (__EXC_STREAM, file, function, line, code);
#endif

  __exc_file = file;
  __exc_function = function;
  __exc_line = line;
  __exc_code = code;

  /* Pop for jumping. */
  __exc_pop (&j);
  
  __exc_debug ("Jumping to the handler");

  /* LONGJUMP to J with nonzero value. */
  longjmp (j, 1);
}

/* Raise it in upper level of catcher blocks. */

static void
__exc_reraise ()
{
  jmp_buf j;

  __exc_debug ("RERAISE ()");
#ifdef __EXC_DEBUG
  __exc_print (__EXC_STREAM, __exc_file, __exc_function,
	       __exc_line, __exc_code);
#endif

  __exc_pop (&j);
  longjmp (j, 1);
}



/* What a f...  Somewhy I can't get GCC's __attribute__ working here
   to tell that FILE and LINE are unused in non-debuging mode. */

/* TODO: define __attribute__ (foo) to do nothing if this is not GCC.
   (Check if GNUC is predefined.) */

static int
__exc_on (char *file, char *function, unsigned line, __EXC_TYPE code)
{
  /* It is impossible to jump to the end of EXCEPT block, nor it is
     possible to mix ELSE with IF, so use before-defined __EXC_HANDLED
     flag. 

     Actually, use of
      on (...) ...;
      else on (...) ...;

     Is possible, but don't force the *user* to care about it. */

  __exc_debug ("ON ()");
  __exc_debug ("Trying to handle in file \"%s\", at line %u", file, line);
#ifdef __EXC_DEBUG
  if (function)
    {
      __exc_debug ("In function \"%s\".", function);
    }
#endif

  if (__exc_handled == 1)
    {
      __exc_debug ("Exception already handled in this level, skip");
      return 0;
    }
  
  if (__EXC_EQ (code, __exc_code))
    {
      __exc_debug ("This handler FITS");

      __exc_handled = 1;
      return 1;
    }

  __exc_debug ("This handler DOESN'T FIT");

  /* Not matched. */
  return 0;
}



/* Start catching. */

/* Obviously, there is no way to check if appropriate EXCEPT exists.
   Its non-exsistence won't do segfault etc.; program will simply do
   the thing after TRY, without any error handling.  Raising from
   there works. */

#define try                                  \
  if (({jmp_buf __exc_j;                     \
        int __exc_ret;                       \
        __exc_ret = setjmp (__exc_j);        \
        __exc_push (&__exc_j, __exc_ret);})) \
    __EXC_END(__exc_pop (0))

#define raise(code...)                       \
  __exc_raise (__FILE__, __EXC_FUNCTION,     \
	       __LINE__, __EXC_MAKE (code))

/* RAISE in EXCEPT block won't go into itself, because corresping item
   from __EXC_GLOBAL was already popped. */

#define except                               \
  else                                       \
    __EXC_BLOCK (__exc_handled = 0,          \
                 ({ if (__exc_handled == 0)  \
                     __exc_reraise (); }))

/* EXPECT is an alias for EXCEPT. */

#define expect                         except

/* CATCH is an alias for EXCEPT. */

#define catch                          except

/* Try to handle an exception. */

#define on(code...)                          \
  if (__exc_on (__FILE__, __EXC_FUNCTION,    \
		__LINE__, __EXC_MAKE (code)))

/* End. */

==> excc-0.1.5.tar.gz: excc-0.1.5/exception.h <==
/* exception.h --- header implementing lightweight, but featured
                   exception handling.

Time-stamp: "2005-04-10 17:06:31 admp"

Copyright (C) 2003-5 Adomas P. <adomas.paltanavicius@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA. */

/* Summary.

To understand all this, you should be experienced in C, and also
know the setjmp interface.  (Unixers do.)

Implemented:
  1. TRY --- a catcher block.
  2. EXCEPT, EXPECT, CATCH --- a handler block.
  3. THROW --- throws an exception.
  4. ON --- a handler.

IMPORTANT:
  TRY without EXCEPT does not work properly.  This is impossible
  to fix, at least I don't see the way.  Anyway, if you don't
  want to break the whole system, I see no other need for
  such a construct.

Syntax:
  1. TRY
  The syntax is:
    try { statements; } except { handler blocks; }

  2. EXCEPT, EXPECT, CATCH
  The syntax is:
    except { handler blocks; }

  3. THROW
  The syntax is: 
    THROW (params)
  PARAMS depend on your local implementation.  By default, it is a
  number (int).
  
  4. ON
  The syntax is:
    on (params) { statements; }
  PARAMS are the same as for THROW.

Supports:
  1. Nested catcher blocks.
  2. Raising an exception outside catcher block (this results in an 
     unhandled exception.)
  3. Raising an exception in function called from catcher block.
  4. Different types for exception structures.
  5. Reporting file and line where exception was thrown/handled. */

/* Search for TODO and FIXME tags if you want to enchance this.
   (This would be nice.) */

/* TODO: since customized exceptions usually allocate memory
   dynamically, it would be nice to have support for freeing
   function, which would be called after it's not needed
   anymore: __EXC_FREE.  I. e. after exception is handled. */

#if !defined __EXCEPTION_H__
#define __EXCEPTION_H__

#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* The default prefix is `__exc' for functions and `__EXC' for
   macros.  Underscores mean you shouldn't touch 'em! */



/* C syntax hacks.

   Oh, god, I felt like an inventor after writing these.
   Clearly, code

     for (start (), J = 1; J; end (), J = 0)
       code ();

   Does this:
   1) Executes START
   2) Executes CODE
   3) Executes END
   4) ...And terminates.

   It also works if nested (think why yourself.)
*/

/* Execute START, then block after the macro, and finally END. */

#define __EXC_BLOCK(start, end)              \
  for (start, __exc_block_pass = 1;          \
       __exc_block_pass;                     \
       end, __exc_block_pass = 0)

/* Likewise, but START block is empty. */

#define __EXC_END(end)                       \
  for (__exc_block_pass = 1;                 \
       __exc_block_pass;                     \
       end, __exc_block_pass = 0)



/* For function name.  GCC includes things which expand to
   the name of current function's name.  */

#if (!defined (__GNUC__) || __GNUC__ < 2 || \
     __GNUC_MINOR__ < (defined (__cplusplus) ? 6 : 4))
   /* Otherwise stick to unknown. */
#  define __EXC_FUNCTION               (char *) 0
#else
#  define __EXC_FUNCTION	       __PRETTY_FUNCTION__
#endif



/* The exception value. */

/* You'll want to make local changes to these.  For example, to use
   your own exception structure. */

/* Exception is by default an int.  Anyway, it can be anything from
   string to some structure.

   Whatever the implementation you choose, type name should be
   defined as __EXC_TYPE.  The THROW (and ON) macro accepts as
   many arguments, s it is given, so your function may use all
   the power of argument passing.  Define your function's name
   as __EXC_MAKE. Exceptions are compared in ON macro.  You
   should define comparing function as __EXC_EQ.

   For example, if you'd like to use strings in place of numbers,
   use this snippet:

   1) #define __EXC_TYPE         char *
   2) #define __EXC_EQ(s1, s2)   (strcasecmp (s1, s2) == 0)
   3) #define __EXC_PRINT(e, stream) \
	         fprintf (stream, "%s", e)

*/

#ifndef __EXC_TYPE
#  define __EXC_TYPE               int

/* Include the default __EXC_PRINT. */
#  define __EXC_TYPE_DEFAULT
#endif

#ifndef __EXC_MAKE
#  define __EXC_MAKE(code...)      code
#endif

#ifndef __EXC_ON
#  define __EXC_ON                 __EXC_MAKE
#endif

#ifndef __EXC_EQ
#  define __EXC_EQ(c1, c2)         ((c1) == (c2))
#endif


/* Optional exception printer.  This is used for debugging purposes
   only.  Define yourself's one as __EXC_PRINT. Arguments are
   exception of type __EXC_TYPE and stream to print to. */

#if !defined (__EXC_PRINT) && defined (__EXC_TYPE_DEFAULT)
#  define __EXC_PRINT(e, stream)             \
     fprintf (stream, "%d", e)
#endif



/* All variables are declared volatile to force non-optimization. 
   They should also be declared as thread-local. */

/* This counter is used by __EXC_BLOCK.  It works well even if nested. */
extern volatile int __exc_block_pass;

/* Flag to be set by ON? */
extern volatile int __exc_handled;

/* For indexing every call to TRY. */
extern volatile unsigned __exc_tries;

/* These identify the thrown exception.  File, function, line and
   the exception itself. */
extern char *__exc_file;
extern char *__exc_function;
extern unsigned __exc_line;
extern volatile __EXC_TYPE __exc_code;

/* Stack is actually a linked list of catcher cells. */
struct __exc_stack
{
  unsigned num;
  jmp_buf j;
  struct __exc_stack *prev;
};

/* This is the global stack of catchers. */
extern struct __exc_stack *__exc_global;



/* Debugging of exceptions.  Nothing interesting for you.  (Just for me.) 
   Anyway, it generates many (really) messages telling what is going
   to happen.  I order to work with it successfully, you should define
   __EXC_PRINT (see above.)
*/
#ifdef __EXC_DEBUG
#  include <stdarg.h>
#  ifndef __EXC_STREAM
/* I often redirect debugging information to a file. */
#    define __EXC_STREAM                     stdout
#  endif

/* Prints error message. */
extern void __exc_debug(char *, ...);

/* For printing __exc_global. */
extern void __exc_print_global(void);

#else
#  define __exc_debug(args...)
#  define __exc_print_global()
#endif



/* Prints information about exception.  Called in debug mode, or when
   no handler is found. */
extern void __exc_print (FILE *, char *, char *,
			 unsigned, __EXC_TYPE);

/* Pop exception from stack, putting into J (if nonzero).  If stack is
   empty, print error message and exit.  Used in EXCEPT. */
extern void __exc_pop (jmp_buf *);



/* Push J onto the stack, with RETURNED as value from SETJMP.  Return
   nonzero, if RETURNED is 0.  If RETURNED is nonzero, returns 0.
   Used in TRY. */
extern int __exc_push (jmp_buf *, int);



/* Throw an exception in FILE at LINE, with code CODE.  Used in THROW. */
extern __attribute__((noreturn)) void __exc_throw (char *, char *, unsigned, __EXC_TYPE);

/* Throw it in upper level of catcher blocks. */
extern void __exc_rethrow ();



/* What a f...  Somewhy I can't get GCC's __attribute__ working here
   to tell that FILE and LINE are unused in non-debuging mode. */

/* TODO: define __attribute__ (foo) to do nothing if this is not GCC.
   (Check if GNUC is predefined.) */

extern int __exc_on (char *, char *, unsigned, __EXC_TYPE);



/* Start catching. */

/* Obviously, there is no way to check if appropriate EXCEPT exists.
   Its non-exsistence won't do segfault etc.; program will simply do
   the thing after TRY, without any error handling.  Raising from
   there works. */

#define try                                  \
  if (({jmp_buf __exc_j;                     \
        int __exc_ret;                       \
        __exc_ret = setjmp (__exc_j);        \
        __exc_push (&__exc_j, __exc_ret);})) \
    __EXC_END(__exc_pop (0))

#define throw(code...)                       \
  __exc_throw (__FILE__, __EXC_FUNCTION,     \
	       __LINE__, __EXC_MAKE (code))

/* THROW in EXCEPT block won't go into itself, because corresping item
   from __EXC_GLOBAL was already popped. */

#define except                               \
  else                                       \
    __EXC_BLOCK (__exc_handled = 0,          \
                 ({ if (__exc_handled == 0)  \
                     __exc_rethrow (); }))

/* EXPECT is an alias for EXCEPT. */

#define expect                         except

/* CATCH is an alias for EXCEPT. */

#define catch                          except

/* Try to handle an exception. */

#define on(code...)                          \
  if (__exc_on (__FILE__, __EXC_FUNCTION,    \
		__LINE__, __EXC_ON (code)))

#endif
/* End. */
